/* gtkemojichooser.c: An Emoji chooser widget
 * Copyright 2017, Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */

#include <gtk/gtk.h>
#include "emoji-chooser.h"

#define BOX_SPACE 6

typedef struct {
  GtkWidget *box;
  GtkWidget *heading;
  GtkWidget *button;
  const char *first;
  gunichar label;
  gboolean empty;
} EmojiSection;

struct _EmojiChooser
{
  GtkBin parent_instance;

  GtkWidget *search_entry;
  GtkWidget *stack;
  GtkWidget *scrolled_window;

  int emoji_max_width;

  EmojiSection recent;
  EmojiSection people;
  EmojiSection body;
  EmojiSection nature;
  EmojiSection food;
  EmojiSection travel;
  EmojiSection activities;
  EmojiSection objects;
  EmojiSection symbols;
  EmojiSection flags;

  GVariant *data;

  GSettings *settings;
};

struct _EmojiChooserClass {
  GtkBinClass parent_class;
};

enum {
  EMOJI_PICKED,
  LAST_SIGNAL
};

static int signals[LAST_SIGNAL];

G_DEFINE_TYPE (EmojiChooser, emoji_chooser, GTK_TYPE_BIN)

static void
emoji_chooser_finalize (GObject *object)
{
  EmojiChooser *chooser = EMOJI_CHOOSER (object);

  g_variant_unref (chooser->data);
  g_object_unref (chooser->settings);

  G_OBJECT_CLASS (emoji_chooser_parent_class)->finalize (object);
}

void
scroll_to_section (GtkButton *button,
                   gpointer   data)
{
  EmojiSection *section = data;
  EmojiChooser *chooser;
  GtkAdjustment *adj;
  GtkAllocation alloc = { 0, 0, 0, 0 };

  chooser = EMOJI_CHOOSER (gtk_widget_get_ancestor (GTK_WIDGET (button), GTK_TYPE_EMOJI_CHOOSER));

  adj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (chooser->scrolled_window));
  if (section->heading)
    gtk_widget_get_allocation (section->heading, &alloc);
  gtk_adjustment_set_value (adj, alloc.y - BOX_SPACE);//FROM: gtk_adjustment_animate_to_value
}

static void
add_emoji (GtkWidget    *box,
           gboolean      prepend,
           GVariant     *item,
           gunichar      modifier,
           EmojiChooser *chooser);

#define MAX_RECENT (7*3)

static void
populate_recent_section (EmojiChooser *chooser)
{
  GVariant *variant;
  GVariant *item;
  GVariantIter iter;
  gboolean empty = FALSE;

  variant = g_settings_get_value (chooser->settings, "recent-emoji");
  g_variant_iter_init (&iter, variant);
  while ((item = g_variant_iter_next_value (&iter)))
    {
      GVariant *emoji_data;
      gunichar modifier;

      emoji_data = g_variant_get_child_value (item, 0);
      g_variant_get_child (item, 1, "u", &modifier);
      add_emoji (chooser->recent.box, FALSE, emoji_data, modifier, chooser);
      g_variant_unref (emoji_data);
      g_variant_unref (item);
      empty = FALSE;
    }

  if (!empty)
    {
      gtk_widget_show (chooser->recent.box);
      gtk_widget_set_sensitive (chooser->recent.button, TRUE);
    }
  g_variant_unref (variant);
}

static void
add_recent_item (EmojiChooser *chooser,
                 GVariant        *item,
                 gunichar         modifier)
{
  GList *children, *l;
  int i;
  GVariantBuilder builder;

  g_variant_ref (item);

  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a((auss)u)"));
  g_variant_builder_add (&builder, "(@(auss)u)", item, modifier);

  children = gtk_container_get_children (GTK_CONTAINER (chooser->recent.box));
  for (l = children, i = 1; l; l = l->next, i++)
    {
      GVariant *item2 = g_object_get_data (G_OBJECT (l->data), "emoji-data");
      gunichar modifier2 = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (l->data), "modifier"));

      if (modifier == modifier2 && g_variant_equal (item, item2))
        {
          gtk_widget_destroy (GTK_WIDGET (l->data));
          i--;
          continue;
        }
      if (i >= MAX_RECENT)
        {
          gtk_widget_destroy (GTK_WIDGET (l->data));
          continue;
        }

      g_variant_builder_add (&builder, "(@(auss)u)", item2, modifier2);
    }
  g_list_free (children);

  add_emoji (chooser->recent.box, TRUE, item, modifier, chooser);

  /* Enable recent */
  gtk_widget_show (chooser->recent.box);
  gtk_widget_set_sensitive (chooser->recent.button, TRUE);

  g_settings_set_value (chooser->settings, "recent-emoji", g_variant_builder_end (&builder));

  g_variant_unref (item);
}

static void
emoji_activated (GtkFlowBox      *box,
                 GtkFlowBoxChild *child,
                 gpointer         data)
{
  EmojiChooser *chooser = data;
  char *text;
  GtkWidget *label;
  GVariant *item;
  gunichar modifier;

  //gtk_popover_popdown (GTK_POPOVER (chooser));

  label = gtk_bin_get_child (GTK_BIN (child));
  text = g_strdup (gtk_label_get_label (GTK_LABEL (label)));

  item = (GVariant*) g_object_get_data (G_OBJECT (child), "emoji-data");
  modifier = (gunichar) GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (child), "modifier"));
  add_recent_item (chooser, item, modifier);

  g_signal_emit (data, signals[EMOJI_PICKED], 0, text);
  g_free (text);
}

static gboolean
has_variations (GVariant *emoji_data)
{
  GVariant *codes;
  int i;
  gboolean has_variations;

  has_variations = FALSE;
  codes = g_variant_get_child_value (emoji_data, 0);
  for (i = 0; i < g_variant_n_children (codes); i++)
    {
      gunichar code;
      g_variant_get_child (codes, i, "u", &code);
      if (code == 0)
        {
          has_variations = TRUE;
          break;
        }
    }
  g_variant_unref (codes);

  return has_variations;
}

static void
show_variations (EmojiChooser *chooser,
                 GtkWidget       *child)
{
  GtkWidget *popover;
  GtkWidget *view;
  GtkWidget *box;
  GVariant *emoji_data;
  //GtkWidget *parent_popover;
  gunichar modifier;

  if (!child)
    return;

  emoji_data = (GVariant*) g_object_get_data (G_OBJECT (child), "emoji-data");
  if (!emoji_data)
    return;

  if (!has_variations (emoji_data))
    return;

  //parent_popover = gtk_widget_get_ancestor (child, GTK_TYPE_POPOVER);
  popover = gtk_popover_new (child);
  view = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
  gtk_style_context_add_class (gtk_widget_get_style_context (view), "view");
  box = gtk_flow_box_new ();
  gtk_flow_box_set_homogeneous (GTK_FLOW_BOX (box), TRUE);
  gtk_flow_box_set_min_children_per_line (GTK_FLOW_BOX (box), 6);
  gtk_flow_box_set_max_children_per_line (GTK_FLOW_BOX (box), 6);
  gtk_flow_box_set_activate_on_single_click (GTK_FLOW_BOX (box), TRUE);
  gtk_flow_box_set_selection_mode (GTK_FLOW_BOX (box), GTK_SELECTION_NONE);
  g_object_set (box, "accept-unpaired-release", TRUE, NULL);
  gtk_container_add (GTK_CONTAINER (popover), view);
  gtk_container_add (GTK_CONTAINER (view), box);

  //g_signal_connect (box, "child-activated", G_CALLBACK (emoji_activated), parent_popover);

  add_emoji (box, FALSE, emoji_data, 0, chooser);
  for (modifier = 0x1f3fb; modifier <= 0x1f3ff; modifier++)
    add_emoji (box, FALSE, emoji_data, modifier, chooser);

  gtk_popover_popup (GTK_POPOVER (popover));
}

static void
long_pressed_cb (GtkGesture *gesture,
                 double      x,
                 double      y,
                 gpointer    data)
{
  EmojiChooser *chooser = data;
  GtkWidget *box;
  GtkWidget *child;

  box = gtk_event_controller_get_widget (GTK_EVENT_CONTROLLER (gesture));
  child = GTK_WIDGET (gtk_flow_box_get_child_at_pos (GTK_FLOW_BOX (box), x, y));
  show_variations (chooser, child);
}

static void
pressed_cb (GtkGesture *gesture,
            int         n_press,
            double      x,
            double      y,
            gpointer    data)
{
  EmojiChooser *chooser = data;
  GtkWidget *box;
  GtkWidget *child;

  box = gtk_event_controller_get_widget (GTK_EVENT_CONTROLLER (gesture));
  child = GTK_WIDGET (gtk_flow_box_get_child_at_pos (GTK_FLOW_BOX (box), x, y));
  show_variations (chooser, child);
}

static gboolean
popup_menu (GtkWidget *widget,
            gpointer   data)
{
  EmojiChooser *chooser = data;

  show_variations (chooser, widget);
  return TRUE;
}

static void
add_emoji (GtkWidget    *box,
           gboolean      prepend,
           GVariant     *item,
           gunichar      modifier,
           EmojiChooser *chooser)
{
  GtkWidget *child;
  GtkWidget *label;
  PangoAttrList *attrs;
  GVariant *codes;
  char text[64];
  char *p = text;
  int i;
  PangoLayout *layout;
  PangoRectangle rect;

  codes = g_variant_get_child_value (item, 0);
  for (i = 0; i < g_variant_n_children (codes); i++)
    {
      gunichar code;

      g_variant_get_child (codes, i, "u", &code);
      if (code == 0)
        code = modifier;
      if (code != 0)
        p += g_unichar_to_utf8 (code, p);
    }
  g_variant_unref (codes);
  p += g_unichar_to_utf8 (0xFE0F, p); /* U+FE0F is the Emoji variation selector */
  p[0] = 0;

  label = gtk_label_new (text);
  attrs = pango_attr_list_new ();
  pango_attr_list_insert (attrs, pango_attr_scale_new (PANGO_SCALE_X_LARGE));
  gtk_label_set_attributes (GTK_LABEL (label), attrs);
  pango_attr_list_unref (attrs);

  layout = gtk_label_get_layout (GTK_LABEL (label));
  pango_layout_get_extents (layout, &rect, NULL);

  /* Check for fallback rendering that generates too wide items */
  if (pango_layout_get_unknown_glyphs_count (layout) > 0 ||
      rect.width >= 1.5 * chooser->emoji_max_width)
    {
      gtk_widget_destroy (label);
      return;
    }

  child = gtk_flow_box_child_new ();
  gtk_style_context_add_class (gtk_widget_get_style_context (child), "emoji");
  g_object_set_data_full (G_OBJECT (child), "emoji-data",
                          g_variant_ref (item),
                          (GDestroyNotify)g_variant_unref);
  if (modifier != 0)
    g_object_set_data (G_OBJECT (child), "modifier", GUINT_TO_POINTER (modifier));

  if (chooser)
    g_signal_connect (child, "popup-menu", G_CALLBACK (popup_menu), chooser);

  gtk_container_add (GTK_CONTAINER (child), label);
  gtk_flow_box_insert (GTK_FLOW_BOX (box), child, prepend ? 0 : -1);
}

static void
populate_emoji_chooser (EmojiChooser *chooser)
{
  GBytes *bytes = NULL;
  GVariantIter iter;
  GVariant *item;
  GtkWidget *box;

  bytes = g_resources_lookup_data ("/org/gtk/libgtk/emoji/emoji.data", 0, NULL);
  chooser->data = g_variant_ref_sink (g_variant_new_from_bytes (G_VARIANT_TYPE ("a(auss)"), bytes, TRUE));

  g_variant_iter_init (&iter, chooser->data);
  box = chooser->people.box;
  while ((item = g_variant_iter_next_value (&iter)))
    {
      const char *name;

      g_variant_get_child (item, 1, "&s", &name);

      if (strcmp (name, chooser->body.first) == 0)
        box = chooser->body.box;
      else if (strcmp (name, chooser->nature.first) == 0)
        box = chooser->nature.box;
      else if (strcmp (name, chooser->food.first) == 0)
        box = chooser->food.box;
      else if (strcmp (name, chooser->travel.first) == 0)
        box = chooser->travel.box;
      else if (strcmp (name, chooser->activities.first) == 0)
        box = chooser->activities.box;
      else if (strcmp (name, chooser->objects.first) == 0)
        box = chooser->objects.box;
      else if (strcmp (name, chooser->symbols.first) == 0)
        box = chooser->symbols.box;
      else if (strcmp (name, chooser->flags.first) == 0)
        box = chooser->flags.box;

      add_emoji (box, FALSE, item, 0, chooser);
      g_variant_unref (item);
    }

  g_bytes_unref (bytes);
}

static void
adj_value_changed (GtkAdjustment *adj,
                   gpointer       data)
{
  EmojiChooser *chooser = data;
  double value = gtk_adjustment_get_value (adj);
  EmojiSection const *sections[] = {
    &chooser->recent,
    &chooser->people,
    &chooser->body,
    &chooser->nature,
    &chooser->food,
    &chooser->travel,
    &chooser->activities,
    &chooser->objects,
    &chooser->symbols,
    &chooser->flags,
  };
  EmojiSection const *select_section = sections[0];
  gsize i;

  /* Figure out which section the current scroll position is within */
  for (i = 0; i < G_N_ELEMENTS (sections); ++i)
    {
      EmojiSection const *section = sections[i];
      GtkAllocation alloc;

      if (section->heading)
        gtk_widget_get_allocation (section->heading, &alloc);
      else
        gtk_widget_get_allocation (section->box, &alloc);

      if (value < alloc.y - BOX_SPACE)
        break;

      select_section = section;
    }

  /* Un/Check the section buttons accordingly */
  for (i = 0; i < G_N_ELEMENTS (sections); ++i)
    {
      EmojiSection const *section = sections[i];

      if (section == select_section)
        gtk_widget_set_state_flags (section->button, GTK_STATE_FLAG_CHECKED, FALSE);
      else
        gtk_widget_unset_state_flags (section->button, GTK_STATE_FLAG_CHECKED);
    }
}

static gboolean
filter_func (GtkFlowBoxChild *child,
             gpointer         data)
{
  EmojiSection *section = data;
  EmojiChooser *chooser;
  GVariant *emoji_data;
  const char *text;
  const char *name;
  gboolean res;

  res = TRUE;

  chooser = EMOJI_CHOOSER (gtk_widget_get_ancestor (GTK_WIDGET (child), GTK_TYPE_EMOJI_CHOOSER));
  text = gtk_entry_get_text (GTK_ENTRY (chooser->search_entry));
  emoji_data = (GVariant *) g_object_get_data (G_OBJECT (child), "emoji-data");

  if (text[0] == 0)
    goto out;

  if (!emoji_data)
    goto out;

  g_variant_get_child (emoji_data, 1, "&s", &name);
  res = g_str_match_string (text, name, TRUE);

out:
  if (res)
    section->empty = FALSE;

  return res;
}

static void
invalidate_section (EmojiSection *section)
{
  section->empty = TRUE;
  gtk_flow_box_invalidate_filter (GTK_FLOW_BOX (section->box));
}

static void
update_headings (EmojiChooser *chooser)
{
  gtk_widget_set_visible (chooser->people.heading, !chooser->people.empty);
  gtk_widget_set_visible (chooser->people.box, !chooser->people.empty);
  gtk_widget_set_visible (chooser->body.heading, !chooser->body.empty);
  gtk_widget_set_visible (chooser->body.box, !chooser->body.empty);
  gtk_widget_set_visible (chooser->nature.heading, !chooser->nature.empty);
  gtk_widget_set_visible (chooser->nature.box, !chooser->nature.empty);
  gtk_widget_set_visible (chooser->food.heading, !chooser->food.empty);
  gtk_widget_set_visible (chooser->food.box, !chooser->food.empty);
  gtk_widget_set_visible (chooser->travel.heading, !chooser->travel.empty);
  gtk_widget_set_visible (chooser->travel.box, !chooser->travel.empty);
  gtk_widget_set_visible (chooser->activities.heading, !chooser->activities.empty);
  gtk_widget_set_visible (chooser->activities.box, !chooser->activities.empty);
  gtk_widget_set_visible (chooser->objects.heading, !chooser->objects.empty);
  gtk_widget_set_visible (chooser->objects.box, !chooser->objects.empty);
  gtk_widget_set_visible (chooser->symbols.heading, !chooser->symbols.empty);
  gtk_widget_set_visible (chooser->symbols.box, !chooser->symbols.empty);
  gtk_widget_set_visible (chooser->flags.heading, !chooser->flags.empty);
  gtk_widget_set_visible (chooser->flags.box, !chooser->flags.empty);

  if (chooser->recent.empty && chooser->people.empty &&
      chooser->body.empty && chooser->nature.empty &&
      chooser->food.empty && chooser->travel.empty &&
      chooser->activities.empty && chooser->objects.empty &&
      chooser->symbols.empty && chooser->flags.empty)
    gtk_stack_set_visible_child_name (GTK_STACK (chooser->stack), "empty");
  else
    gtk_stack_set_visible_child_name (GTK_STACK (chooser->stack), "list");
}

static void
search_changed (GtkEntry *entry,
                gpointer  data)
{
  EmojiChooser *chooser = data;

  invalidate_section (&chooser->recent);
  invalidate_section (&chooser->people);
  invalidate_section (&chooser->body);
  invalidate_section (&chooser->nature);
  invalidate_section (&chooser->food);
  invalidate_section (&chooser->travel);
  invalidate_section (&chooser->activities);
  invalidate_section (&chooser->objects);
  invalidate_section (&chooser->symbols);
  invalidate_section (&chooser->flags);

  update_headings (chooser);
}

static void
setup_section (EmojiChooser *chooser,
               EmojiSection   *section,
               const char     *first,
               const char     *icon)
{
  GtkAdjustment *adj;

  section->first = first;

  GtkWidget* image = gtk_image_new_from_icon_name (icon, GTK_ICON_SIZE_SMALL_TOOLBAR);
  gtk_button_set_image (GTK_BUTTON (section->button), image);

  adj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (chooser->scrolled_window));

  //gtk_container_set_focus_vadjustment (GTK_CONTAINER (section->box), adj);
  gtk_flow_box_set_filter_func (GTK_FLOW_BOX (section->box), filter_func, section, NULL);
  g_signal_connect (section->button, "clicked", G_CALLBACK (scroll_to_section), section);
}

static void
emoji_chooser_init (EmojiChooser *chooser)
{
  GtkAdjustment *adj;

  chooser->settings = g_settings_new ("org.gtk.Settings.EmojiChooser");

  gtk_widget_init_template (GTK_WIDGET (chooser));

  /* Get a reasonable maximum width for an emoji. We do this to
   * skip overly wide fallback rendering for certain emojis the
   * font does not contain and therefore end up being rendered
   * as multiply glyphs.
   */
  {
    PangoLayout *layout = gtk_widget_create_pango_layout (GTK_WIDGET (chooser), "🙂");
    PangoAttrList *attrs;
    PangoRectangle rect;

    attrs = pango_attr_list_new ();
    pango_attr_list_insert (attrs, pango_attr_scale_new (PANGO_SCALE_X_LARGE));
    pango_layout_set_attributes (layout, attrs);
    pango_attr_list_unref (attrs);

    pango_layout_get_extents (layout, &rect, NULL);
    chooser->emoji_max_width = rect.width;

    g_object_unref (layout);
  }

  adj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (chooser->scrolled_window));
  g_signal_connect (adj, "value-changed", G_CALLBACK (adj_value_changed), chooser);

  setup_section (chooser, &chooser->recent, NULL, "emoji-recent-symbolic");
  setup_section (chooser, &chooser->people, "grinning face", "emoji-people-symbolic");
  setup_section (chooser, &chooser->body, "selfie", "emoji-body-symbolic");
  setup_section (chooser, &chooser->nature, "monkey face", "emoji-nature-symbolic");
  setup_section (chooser, &chooser->food, "grapes", "emoji-food-symbolic");
  setup_section (chooser, &chooser->travel, "globe showing Europe-Africa", "emoji-travel-symbolic");
  setup_section (chooser, &chooser->activities, "jack-o-lantern", "emoji-activities-symbolic");
  setup_section (chooser, &chooser->objects, "muted speaker", "emoji-objects-symbolic");
  setup_section (chooser, &chooser->symbols, "ATM sign", "emoji-symbols-symbolic");
  setup_section (chooser, &chooser->flags, "chequered flag", "emoji-flags-symbolic");

  populate_emoji_chooser (chooser);
  populate_recent_section (chooser);

  /* We scroll to the top on show, so check the right button for the 1st time */
  gtk_widget_set_state_flags (chooser->recent.button, GTK_STATE_FLAG_CHECKED, FALSE);
}

static void
emoji_chooser_show (GtkWidget *widget)
{
  EmojiChooser *chooser = EMOJI_CHOOSER (widget);
  GtkAdjustment *adj;

  GTK_WIDGET_CLASS (emoji_chooser_parent_class)->show (widget);

  adj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (chooser->scrolled_window));
  gtk_adjustment_set_value (adj, 0);

  gtk_entry_set_text (GTK_ENTRY (chooser->search_entry), "");
}

static void
emoji_chooser_class_init (EmojiChooserClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->finalize = emoji_chooser_finalize;
  widget_class->show = emoji_chooser_show;

  signals[EMOJI_PICKED] = g_signal_new ("emoji-picked",
                                        G_OBJECT_CLASS_TYPE (object_class),
                                        G_SIGNAL_RUN_LAST,
                                        0,
                                        NULL, NULL,
                                        NULL,
                                        G_TYPE_NONE, 1, G_TYPE_STRING|G_SIGNAL_TYPE_STATIC_SCOPE);

gchar* s = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
"<interface domain=\"gtk30\">"
"<template class=\"EmojiChooser\" parent=\"GtkBin\">"
"<style>"
"<class name=\"emoji-picker\"/>"
"</style>"
"<child>"
"<object class=\"GtkBox\" id=\"box\">"
"<property name=\"orientation\">vertical</property>"
"<property name=\"visible\">1</property>"
"<child>"
"<object class=\"GtkSearchEntry\" id=\"search_entry\">"
"<property name=\"visible\">1</property>"
"<property name=\"input-hints\">no-emoji</property>"
"<signal name=\"search-changed\" handler=\"search_changed\"/>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkStack\" id=\"stack\">"
"<property name=\"visible\">1</property>"
"<child>"
"<object class=\"GtkBox\">"
"<property name=\"visible\">1</property>"
"<property name=\"orientation\">vertical</property>"
"<child>"
"<object class=\"GtkScrolledWindow\" id=\"scrolled_window\">"
"<property name=\"visible\">1</property>"
"<property name=\"vexpand\">1</property>"
"<property name=\"hscrollbar-policy\">never</property>"
"<property name=\"min-content-height\">250</property>"
"<style>"
"<class name=\"view\"/>"
"</style>"
"<child>"
"<object class=\"GtkBox\" id=\"emoji_box\">"
"<property name=\"visible\">1</property>"
"<property name=\"orientation\">vertical</property>"
"<property name=\"margin\">6</property>"
"<property name=\"spacing\">6</property>"
"<!-- Remember to keep headings here in sync with button tooltips below -->"
"<child>"
"<object class=\"GtkFlowBox\" id=\"recent.box\">"
"<property name=\"visible\">1</property>"
"<property name=\"homogeneous\">1</property>"
"<property name=\"selection-mode\">none</property>"
"<signal name=\"child-activated\" handler=\"emoji_activated\"/>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkLabel\" id=\"people.heading\">"
"<property name=\"visible\">1</property>"
"<property name=\"label\" translatable=\"yes\">Smileys &amp; People</property>"
"<property name=\"xalign\">0</property>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkFlowBox\" id=\"people.box\">"
"<property name=\"visible\">1</property>"
"<property name=\"homogeneous\">1</property>"
"<property name=\"selection-mode\">none</property>"
"<signal name=\"child-activated\" handler=\"emoji_activated\"/>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkLabel\" id=\"body.heading\">"
"<property name=\"visible\">1</property>"
"<property name=\"label\" translatable=\"yes\">Body &amp; Clothing</property>"
"<property name=\"xalign\">0</property>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkFlowBox\" id=\"body.box\">"
"<property name=\"visible\">1</property>"
"<property name=\"homogeneous\">1</property>"
"<property name=\"selection-mode\">none</property>"
"<signal name=\"child-activated\" handler=\"emoji_activated\"/>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkLabel\" id=\"nature.heading\">"
"<property name=\"visible\">1</property>"
"<property name=\"label\" translatable=\"yes\">Animals &amp; Nature</property>"
"<property name=\"xalign\">0</property>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkFlowBox\" id=\"nature.box\">"
"<property name=\"visible\">1</property>"
"<property name=\"homogeneous\">1</property>"
"<property name=\"selection-mode\">none</property>"
"<signal name=\"child-activated\" handler=\"emoji_activated\"/>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkLabel\" id=\"food.heading\">"
"<property name=\"visible\">1</property>"
"<property name=\"label\" translatable=\"yes\">Food &amp; Drink</property>"
"<property name=\"xalign\">0</property>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkFlowBox\" id=\"food.box\">"
"<property name=\"visible\">1</property>"
"<property name=\"homogeneous\">1</property>"
"<property name=\"selection-mode\">none</property>"
"<signal name=\"child-activated\" handler=\"emoji_activated\"/>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkLabel\" id=\"travel.heading\">"
"<property name=\"visible\">1</property>"
"<property name=\"label\" translatable=\"yes\">Travel &amp; Places</property>"
"<property name=\"xalign\">0</property>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkFlowBox\" id=\"travel.box\">"
"<property name=\"visible\">1</property>"
"<property name=\"homogeneous\">1</property>"
"<property name=\"selection-mode\">none</property>"
"<signal name=\"child-activated\" handler=\"emoji_activated\"/>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkLabel\" id=\"activities.heading\">"
"<property name=\"visible\">1</property>"
"<property name=\"label\" translatable=\"yes\">Activities</property>"
"<property name=\"xalign\">0</property>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkFlowBox\" id=\"activities.box\">"
"<property name=\"visible\">1</property>"
"<property name=\"homogeneous\">1</property>"
"<property name=\"selection-mode\">none</property>"
"<signal name=\"child-activated\" handler=\"emoji_activated\"/>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkLabel\" id=\"objects.heading\">"
"<property name=\"visible\">1</property>"
"<property name=\"label\" translatable=\"yes\" context=\"emoji category\">Objects</property>"
"<property name=\"xalign\">0</property>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkFlowBox\" id=\"objects.box\">"
"<property name=\"visible\">1</property>"
"<property name=\"homogeneous\">1</property>"
"<property name=\"selection-mode\">none</property>"
"<signal name=\"child-activated\" handler=\"emoji_activated\"/>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkLabel\" id=\"symbols.heading\">"
"<property name=\"visible\">1</property>"
"<property name=\"label\" translatable=\"yes\">Symbols</property>"
"<property name=\"xalign\">0</property>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkFlowBox\" id=\"symbols.box\">"
"<property name=\"visible\">1</property>"
"<property name=\"homogeneous\">1</property>"
"<property name=\"selection-mode\">none</property>"
"<signal name=\"child-activated\" handler=\"emoji_activated\"/>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkLabel\" id=\"flags.heading\">"
"<property name=\"visible\">1</property>"
"<property name=\"label\" translatable=\"yes\">Flags</property>"
"<property name=\"xalign\">0</property>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkFlowBox\" id=\"flags.box\">"
"<property name=\"visible\">1</property>"
"<property name=\"homogeneous\">1</property>"
"<property name=\"selection-mode\">none</property>"
"<signal name=\"child-activated\" handler=\"emoji_activated\"/>"
"</object>"
"</child>"
"</object>"
"</child>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkBox\">"
"<property name=\"visible\">1</property>"
"<!-- Remember to keep tooltips here in sync with section headings above -->"
"<child>"
"<object class=\"GtkButton\" id=\"recent.button\">"
"<property name=\"visible\">1</property>"
"<property name=\"relief\">none</property>"
"<property name=\"tooltip-text\" translatable=\"yes\">Recent</property>"
"<style>"
"<class name=\"emoji-section\"/>"
"</style>"
"<child>"
"<object class=\"GtkImage\" id=\"recent.icon\">"
"<property name=\"visible\">1</property>"
"</object>"
"</child>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkButton\" id=\"people.button\">"
"<property name=\"visible\">1</property>"
"<property name=\"relief\">none</property>"
"<property name=\"tooltip-text\" translatable=\"yes\">Smileys &amp; People</property>"
"<style>"
"<class name=\"emoji-section\"/>"
"</style>"
"<child>"
"<object class=\"GtkImage\" id=\"people.icon\">"
"<property name=\"visible\">1</property>"
"</object>"
"</child>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkButton\" id=\"body.button\">"
"<property name=\"visible\">1</property>"
"<property name=\"relief\">none</property>"
"<property name=\"tooltip-text\" translatable=\"yes\">Body &amp; Clothing</property>"
"<style>"
"<class name=\"emoji-section\"/>"
"</style>"
"<child>"
"<object class=\"GtkImage\" id=\"body.icon\">"
"<property name=\"visible\">1</property>"
"</object>"
"</child>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkButton\" id=\"nature.button\">"
"<property name=\"visible\">1</property>"
"<property name=\"relief\">none</property>"
"<property name=\"tooltip-text\" translatable=\"yes\">Animals &amp; Nature</property>"
"<style>"
"<class name=\"emoji-section\"/>"
"</style>"
"<child>"
"<object class=\"GtkImage\" id=\"nature.icon\">"
"<property name=\"visible\">1</property>"
"</object>"
"</child>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkButton\" id=\"food.button\">"
"<property name=\"visible\">1</property>"
"<property name=\"relief\">none</property>"
"<property name=\"tooltip-text\" translatable=\"yes\">Food &amp; Drink</property>"
"<style>"
"<class name=\"emoji-section\"/>"
"</style>"
"<child>"
"<object class=\"GtkImage\" id=\"food.icon\">"
"<property name=\"visible\">1</property>"
"</object>"
"</child>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkButton\" id=\"travel.button\">"
"<property name=\"visible\">1</property>"
"<property name=\"relief\">none</property>"
"<property name=\"tooltip-text\" translatable=\"yes\">Travel &amp; Places</property>"
"<style>"
"<class name=\"emoji-section\"/>"
"</style>"
"<child>"
"<object class=\"GtkImage\" id=\"travel.icon\">"
"<property name=\"visible\">1</property>"
"</object>"
"</child>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkButton\" id=\"activities.button\">"
"<property name=\"visible\">1</property>"
"<property name=\"relief\">none</property>"
"<property name=\"tooltip-text\" translatable=\"yes\">Activities</property>"
"<style>"
"<class name=\"emoji-section\"/>"
"</style>"
"<child>"
"<object class=\"GtkImage\" id=\"activities.icon\">"
"<property name=\"visible\">1</property>"
"</object>"
"</child>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkButton\" id=\"objects.button\">"
"<property name=\"visible\">1</property>"
"<property name=\"relief\">none</property>"
"<property name=\"tooltip-text\" translatable=\"yes\" context=\"emoji category\">Objects</property>"
"<style>"
"<class name=\"emoji-section\"/>"
"</style>"
"<child>"
"<object class=\"GtkImage\" id=\"objects.icon\">"
"<property name=\"visible\">1</property>"
"</object>"
"</child>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkButton\" id=\"symbols.button\">"
"<property name=\"visible\">1</property>"
"<property name=\"relief\">none</property>"
"<property name=\"tooltip-text\" translatable=\"yes\">Symbols</property>"
"<style>"
"<class name=\"emoji-section\"/>"
"</style>"
"<child>"
"<object class=\"GtkImage\" id=\"symbols.icon\">"
"<property name=\"visible\">1</property>"
"</object>"
"</child>"
"</object>"
"</child>"
"<child>"
"<object class=\"GtkButton\" id=\"flags.button\">"
"<property name=\"visible\">1</property>"
"<property name=\"relief\">none</property>"
"<property name=\"tooltip-text\" translatable=\"yes\">Flags</property>"
"<style>"
"<class name=\"emoji-section\"/>"
"</style>"
"<child>"
"<object class=\"GtkImage\" id=\"flags.icon\">"
"<property name=\"visible\">1</property>"
"</object>"
"</child>"
"</object>"
"</child>"
"</object>"
"</child>"
"</object>"
"<packing>"
"<property name=\"name\">list</property>"
"</packing>"
"</child>"
"<child>"
"<object class=\"GtkGrid\">"
"<property name=\"visible\">1</property>"
"<property name=\"row-spacing\">12</property>"
"<property name=\"halign\">center</property>"
"<property name=\"valign\">center</property>"
"<style>"
"<class name=\"dim-label\"/>"
"</style>"
"<child>"
"<object class=\"GtkImage\">"
"<property name=\"visible\">1</property>"
"<property name=\"icon-name\">edit-find-symbolic</property>"
"<property name=\"pixel-size\">72</property>"
"<style>"
"<class name=\"dim-label\"/>"
"</style>"
"</object>"
"<packing>"
"<property name=\"left-attach\">0</property>"
"<property name=\"top-attach\">0</property>"
"</packing>"
"</child>"
"<child>"
"<object class=\"GtkLabel\">"
"<property name=\"visible\">1</property>"
"<property name=\"label\" translatable=\"yes\">No Results Found</property>"
"<attributes>"
"<attribute name=\"weight\" value=\"bold\"/>"
"<attribute name=\"scale\" value=\"1.44\"/>"
"</attributes>"
"</object>"
"<packing>"
"<property name=\"left-attach\">0</property>"
"<property name=\"top-attach\">1</property>"
"</packing>"
"</child>"
"<child>"
"<object class=\"GtkLabel\">"
"<property name=\"visible\">1</property>"
"<property name=\"label\" translatable=\"yes\">Try a different search</property>"
"<style>"
"<class name=\"dim-label\"/>"
"</style>"
"</object>"
"<packing>"
"<property name=\"left-attach\">0</property>"
"<property name=\"top-attach\">2</property>"
"</packing>"
"</child>"
"</object>"
"<packing>"
"<property name=\"name\">empty</property>"
"</packing>"
"</child>"
"</object>"
"</child>"
"</object>"
"</child>"
"</template>"
"</interface>";
  gtk_widget_class_set_template (widget_class, g_bytes_new_static(s, strlen(s)));

  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, search_entry);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, stack);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, scrolled_window);

  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, recent.box);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, recent.button);

  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, people.box);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, people.heading);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, people.button);

  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, body.box);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, body.heading);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, body.button);

  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, nature.box);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, nature.heading);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, nature.button);

  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, food.box);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, food.heading);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, food.button);

  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, travel.box);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, travel.heading);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, travel.button);

  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, activities.box);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, activities.heading);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, activities.button);

  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, objects.box);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, objects.heading);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, objects.button);

  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, symbols.box);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, symbols.heading);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, symbols.button);

  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, flags.box);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, flags.heading);
  gtk_widget_class_bind_template_child (widget_class, EmojiChooser, flags.button);

  gtk_widget_class_bind_template_callback (widget_class, emoji_activated);
  gtk_widget_class_bind_template_callback (widget_class, search_changed);
  gtk_widget_class_bind_template_callback (widget_class, pressed_cb);
  gtk_widget_class_bind_template_callback (widget_class, long_pressed_cb);
}

GtkWidget *
emoji_chooser_new (void)
{
  return GTK_WIDGET (g_object_new (GTK_TYPE_EMOJI_CHOOSER, NULL));
}
